use rocket::form::Form; // for creating form struct
use rocket::State;
use rocket::fs::NamedFile;
use sqlx::mysql::MySqlPool;
use std::fs::File;
use std::io::prelude::*;
use rocket::mtls::Certificate;
use rocket::response::content::RawHtml;
use rocket::http::ContentType;

#[derive(FromForm)]
pub struct RegisterForm {
    username: String,
    password: String,
    confirmpassword: String,
}

#[derive(FromForm)]
pub struct LoginForm {
    username: String,
    password: String,
    otp: String,
}

pub struct Pool(pub MySqlPool);

/* This function takes the user's submitted registration form and adds them to the users table of
 * the database if the user's password and confirmation password are equal. If there are any
 * errors during the database operation, the error is safely handled. */
#[post("/user/create", data = "<register_form>")]
pub async fn user_create(register_form: Form<RegisterForm>, pool: &State<Pool>) -> RawHtml<String>
{
    if !register_form.password.eq(&register_form.confirmpassword) // if the password and the
                                                                  // confirmation password aren't
                                                                  // the same
    {
        let response = format!("Password and confirmation password are not the same.");
        RawHtml(response) // return error to user
    }
    else
    {
        let hash_result = super::database::hash_pass(&register_form.password).await; // hash password
        let hash = match hash_result { // if the function returned a hash, then save it, if it
                                       // returned an error then show the error to the user
            Ok(hash) => hash,
            Err(e) => {
                return RawHtml(format!("Error hashing password, {}", e)) // return error to the user
            }
        };
        
        // add user
        let uuid_result = super::database::add_user(&register_form.username, &hash, &pool.0).await;
    
        let uuid = match uuid_result { // if the function returned an id, then save it, if it returned an error
                          // then show the error to the user
            Ok(uuid) => uuid,
            Err(e) => return RawHtml(format!("Error adding user, {}", e)) // return error to the user
        };

        let otp_result = super::database::create_otp(&uuid, &pool.0).await;

        let otp = match otp_result {
            Ok(otp) => otp,
            Err(e) => return RawHtml(format!("Error generating OTP, {}", e))
        };

        let response = format!("<html><body><p>Your OTP secret is: {}</p>\n<p>Please click <a href='/user/{}/cert.p12'>here</a> to download your p12 file.</p></body</html>", otp, uuid);

        RawHtml(response)
    }
}

/* This function is absolutely bloody awful right now but I have no idea how to generate a
 * reference to a file without actually writing to the hard disk :( */
#[get("/user/<uuid>/cert.p12")]
pub async fn user_cert(uuid: String, pool: &State<Pool>) -> (ContentType, NamedFile)
{
    let cert_result = super::database::create_cert(&uuid, &pool.0).await;

    let cert = match cert_result {
        Ok(cert) => cert,
        Err(e) => panic!("Error creating cert, {}", e)
    };

    let cert_file_result = File::create("/tmp/browser.p12");

    let mut cert_file = match cert_file_result {
        Ok(cert_file) => cert_file,
        Err(e) => panic!("Error creating cert file, {}", e)
    };

    let cert_der = match cert.to_der() {
        Ok(cert_der) => cert_der,
        Err(e) => panic!("Error converting cert to der, {}", e)
    };

    let write_result = cert_file.write_all(&cert_der[..]);

    match write_result {
        Ok(_w) => _w,
        Err(e) => panic!("Error writing to cert file, {}", e)
    };

    let named_cert_result = NamedFile::open("/tmp/browser.p12").await;

    let named_cert = match named_cert_result {
        Ok(named_cert) => named_cert,
        Err(e) => panic!("Error converting cert to der, {}", e)
    };

    let content_type = ContentType::new("application", "x-pkcs12");
    (content_type, named_cert)
}

#[post("/user/login", data = "<login_form>")]
pub async fn user_login(_cert: Certificate<'_>, login_form: Form<LoginForm>, pool: &State<Pool>) -> String
{
    let hash_result = super::database::retrieve_hash(&login_form.username, &pool.0).await;

    let hash_result_value = match hash_result {
        Ok(hash_result_value) => hash_result_value,
        Err(e) => return format!("Error, username doesn't exist: {}", e)
    };

    let comparison_result = super::database::verify_pass(&login_form.password, &hash_result_value).await;

    let comparison_result_value = match comparison_result {
        Ok(comparison_result_value) => comparison_result_value,
        Err(e) => return format!("Error when comparing hash and password: {}", e)
    };

    let otp_result = super::database::verify_otp(&login_form.username, &login_form.otp, &pool.0).await;

    let otp_result_value = match otp_result {
        Ok(result_value) => result_value,
        Err(e) => return format!("Error when verifying otp: {}", e)
    };

    if otp_result_value && comparison_result_value
    {
        return format!("You've successfully authenticated!");
    }
    else
    {
        return format!("Incorrect login.");
    }
}

#[cfg(test)]

mod main_tests {
    use super::super::rocket;
    use rocket::local::asynchronous::Client;
    use rocket::http::Status;
    use rocket::http::ContentType;

    /* This unit test checks if the /register route outputs that the password and confirmation
     * password are not equal when provided with two different inputs for those fields. */
    #[rocket::async_test]
    async fn register_test() {
        let client = Client::tracked(super::super::rocket().await).await.expect("Valid instance");
        let response = client
            .post("/register")
            .body("username=test&password=1&confirmpassword=2")
            .header(ContentType::Form) 
            .dispatch()
            .await;
        assert_eq!(response.status(), Status::Ok);
        assert_eq!(response.into_string().await.unwrap(), "Password and confirmation password are not equal.");
    }

    /* This unit test checks if the connection to the database works properly */
    #[rocket::async_test]
    async fn pool_test() {
        let pool_result = super::super::database::connect().await;    
        match pool_result {
            Ok(_p) => assert!(true),
            Err(_e) => assert!(false)
        };
    }
}
