use sqlx::mysql::MySqlPool; // for connecting to db
use argon2::Argon2;
use std::{env, fs};
use argon2::password_hash::{PasswordHash, PasswordHasher, PasswordVerifier, SaltString};
use rand::rngs::OsRng;
use rand::Rng;
use openssl::error::ErrorStack;
use openssl::pkey::PKey;
use openssl::x509::X509;
use totp_rs::{Algorithm, TOTP, Secret};
use openssl::pkcs12::Pkcs12;
use data_encoding::base32;

pub async fn connect() -> Result<MySqlPool, sqlx::Error>
{
    let url = env!("DATABASE_URL");

    let pool = MySqlPool::connect(url).await?;

    Ok(pool)
}

pub async fn hash_pass(password: &String) -> Result<String, argon2::password_hash::Error> {
    let argon2 = Argon2::default(); // init new instance of argon2 hashing
    let password_u8 = password.as_bytes(); // get value of password as bytes
    let salt = SaltString::generate(&mut OsRng); // generate salt
    let hash = argon2.hash_password(password_u8, &salt)?.to_string(); // generate hash of password using salt

    Ok(hash) // return hash
}

pub async fn verify_pass(password: &String, hash: &String) -> Result<bool, argon2::password_hash::Error> {
    let argon2 = Argon2::default();
    let password_u8 = password.as_bytes();
    let hash_parsed = PasswordHash::new(hash)?;
    let result = argon2.verify_password(password_u8, &hash_parsed).is_ok();
    if result
    {
        return Ok(true);
    }
    else
    {
        return Ok(false);
    }
}

pub async fn retrieve_hash(username: &String, pool: &MySqlPool) -> Result<String, sqlx::Error>
{
    let result = sqlx::query!(
        r#"
        select password from users
        where username = ?
        "#,
        username)
        .fetch_one(pool)
        .await?;

    let hash: String = result.password.unwrap();

    Ok(hash)
}

pub async fn add_user(username: &String, hash: &String, pool: &MySqlPool) -> Result<String, sqlx::Error>
{
    sqlx::query!(
        r#"
        insert into users ( username, password )
        values ( ?, ? )
        "#,
        username,
        hash 
        )
        .execute(pool)
        .await?;

    let user_id_result = sqlx::query!(
        r#"
        select id from users
        where username=?
        "#,
        username
        )
        .fetch_one(pool)
        .await?;

    Ok(user_id_result.id)
}

pub async fn create_otp(uuid: &String, pool: &MySqlPool) -> Result<String, sqlx::Error>
{
    let mut secret = String::from("");
    for _ in 1..22
    {
        let mut rng = rand::thread_rng();
        let letter: char = rng.gen_range(b'a'..b'z') as char;
        secret.push(letter);
    }

    let base32_secret = base32::encode(secret.as_bytes());
    let trimmed_secret = base32_secret.replace("=", ""); // trim the equals from the base32 encoded
                                                         // string

    sqlx::query!(
        r#"
        update users
        set otpSecret = ?
        where id = ?
        "#,
        trimmed_secret, // store base32 encoded secret
        uuid
        )
        .execute(pool)
        .await?
        .rows_affected();

    return Ok(trimmed_secret);
}

pub async fn verify_otp(username: &String, otp_code: &String, pool: &MySqlPool) -> Result<bool, sqlx::Error>
{
    let otp_result = sqlx::query!(
        r#"
        select otpSecret from users
        where username = ?
        "#,
        username
        )
        .fetch_one(pool)
        .await?;

    let otp_secret: String = otp_result.otpSecret.unwrap();

    let totp = TOTP::new(
        Algorithm::SHA1,
        6,
        1,
        30,
        Secret::Encoded(otp_secret.to_string()).to_bytes().unwrap(),
    ).unwrap();
    let expected_code = totp.generate_current().unwrap();

    if otp_code.eq(&expected_code)
    {
        return Ok(true)
    }
    else
    {
        return Ok(false)
    }
}

pub async fn create_cert(uuid: &String, pool: &MySqlPool) -> Result<Pkcs12, ErrorStack>
{
    let ca_crt_pem_result = fs::read_to_string("tls/ca_crt.pem");

    let ca_crt_pem = match ca_crt_pem_result {
        Ok(ca_crt_pem) => ca_crt_pem,
        Err(e) => panic!("{}", e)
    };

    let ca_key_pem_result = fs::read_to_string("tls/ca_key.pem");

    let ca_key_pem = match ca_key_pem_result {
        Ok(ca_key_pem) => ca_key_pem,
        Err(e) => panic!("{}", e)
    };

    let ca_crt_pem_obj = X509::from_pem(ca_crt_pem.as_bytes())?;

    let ca_key_pem_obj = PKey::private_key_from_pem(ca_key_pem.as_bytes())?;

    let cert_and_key = super::cert::create_ca_signed_cert(&ca_crt_pem_obj, &ca_key_pem_obj).await?;

    let cert = cert_and_key.0;
    let key = cert_and_key.1;

    let pkcs12_keystore = super::cert::create_pkcs12(&cert, &key).await?;

    let pkcs12_der = match pkcs12_keystore.to_der() {
        Ok(pkcs12_der) => pkcs12_der,
        Err(e) => panic!("{}", e)
    };

    let cert_id_result = sqlx::query!(
        r#"
        insert into certs ( userID, pkcs12 )
        values ( ?, ? )
        "#,
        uuid,
        &pkcs12_der[..]
        )
        .execute(pool)
        .await;

    let _o = match cert_id_result {
        Ok(_o) => _o,
        Err(e) => panic!("{}", e)
    };

    Ok(pkcs12_keystore)
}

#[cfg(test)]
mod database_tests {
    use sqlx::mysql::MySqlPool;

    /* This unit test checks if the add user functionality successfully inserts the record
     * into the database */
    #[sqlx::test]
    async fn add_user_test(pool: MySqlPool) {
        use super::*;
        let username = String::from("test");
        let password = String::from("test");
        let id_result = add_user(&username, &password, &pool).await; // add test user to database
        match id_result { // panic if add user returns an error
            Ok(_p) => assert!(true),
            Err(_e) => assert!(false)
        };
    }

    /* This unit test checks if the connection to the database works properly */
    #[rocket::async_test]
    async fn pool_test() {
        use super::*;
        let pool_result = connect().await; // connect to the database
        match pool_result { // panic if connect returns an error
            Ok(_p) => assert!(true),
            Err(_e) => assert!(false)
        };
    }
}
